//test measurement sequentially, reading an increasing number of channels
// THE COMMENTED SECTIONS represent the code lines to printout the corresponding binary value
#include <Wire.h>
//#include <dummy.h>
#include <stdint.h>
#define SLAVE 0x50
#define NUM_SENSORS 4
int OFFSET = 0;
int capdac = round(OFFSET / 3.125);
int neg =0; //variable used to set if my values is positive or negative and print out the current values respectively
double data_list[4] = {};
int16_t check1[2];
int16_t check_1[4];
float Coffset = capdac * 3.125;
String capdac_hex = String(capdac, HEX);

void setup() {
  delay(1000);
  Wire.begin(21, 22);
  Serial.begin(9600);

  //reset the first FDC
  Wire.beginTransmission(SLAVE);
  Wire.write(0x0C);
  Wire.write(0x80);
  Wire.endTransmission();

  uint16_t register_value = 0x1000 + (capdac << 5);
  uint8_t reg_highbyte = (register_value >> 8) & 0xFF;
  uint8_t reg_lowbyte = register_value & 0xFF;
 
  //Initiate CIN 1-4 for single ended measurement with offset of 15.625 pF
  for (int i = 0; i < 4; i++) {
    Wire.beginTransmission(SLAVE);  
    Wire.write(0x08 + i);          
    Wire.write(reg_highbyte + 32 * i);
    Wire.write(reg_lowbyte);
    Wire.endTransmission();
  }
  
  for (int i = 0; i < NUM_SENSORS; i++) {
    Wire.beginTransmission(SLAVE);
    Wire.write(0x08 + i);
    Wire.endTransmission();
    Wire.requestFrom(SLAVE, 2);
    for (int j = 0; j < 2; j++) {
      check1[j] = Wire.read();
    }
    check_1[i] = (check1[0] << 8) + check1[1];
  }

  //Initiate measurements with data rate of 100S/s and continuous measurements
  Wire.beginTransmission(SLAVE);
  Wire.write(0x0C);  //address FDC register description
  Wire.write(0x05);  //set velocity at 100 S/s --> bits [11:10]
  Wire.write(0xF0);
  Wire.endTransmission();


  //Calibration 
  if (isReady(0)) { 
    data_list[0] = readC(0, &neg); //Read MEAS1 
  }
  if (isReady(1)) { 
    data_list[1] = readC(1, &neg); //Read MEAS2
  }
  if (isReady(2)) { 
    data_list[2] = readC(2, &neg); //Read MEAS3
  }
  if (isReady(3)) { 
    data_list[3] = readC(3, &neg); //Read MEAS4
  }

  // during calibration the system waits and reset the arduino until the channels are correclty read and associated to the sensors
  // if either of the read values is 0, it means there is an error and I have to read again the whole set of measurements
  while (data_list[0] == 0 || data_list[1] == 0 || data_list[2] == 0 || data_list[3] == 0) { 
    if (isReady(0)) {
      data_list[0] = readC(0, &neg);
    }
    if (isReady(1)) {
      data_list[1] = readC(1, &neg);
    }
    if (isReady(2)) {
      data_list[2] = readC(2, &neg);
    }
    if (isReady(3)) {
      data_list[3] = readC(3, &neg);
    }
  }
}

void loop() {
  Serial.println("");
  if (isReady(0)) {
    data_list[0] = readC(0, &neg);
  }
  if (isReady(1)) {
    data_list[1] = readC(1, &neg);
  }
  if (isReady(2)) {
    data_list[2] = readC(2, &neg);
  }
  if (isReady(3)) {
    data_list[3] = readC(3, &neg);
  }
  for (int i = 0; i < 4; i++) {
    if (neg == 1){
      Serial.print("CH" + String(i+1) + ": -" + String(data_list[i], 5) + " " ); //negative value
    } else if (neg == 0){
      Serial.print("CH" + String(i+1) + ": +" + String(data_list[i], 5) + " " ); //positive value
    }
    Serial.println("neg flag is :" + String(neg));
  }
  Serial.println("");
}


bool isReady(int ch_n) {
  int crash = 0;
  crash++;
  if (crash > 20) Serial.println("error");
  int fdc[4];
  int i = 0;
  Wire.beginTransmission(SLAVE);
  Wire.write(0x0C);  
  Wire.endTransmission();

  Wire.requestFrom(SLAVE, 4);  

  while (Wire.available()) {
    fdc[i] = Wire.read();
    if (crash > 20) Serial.print(fdc[i], BIN);
    i++;
  }
  return bitRead(fdc[1], 3 - ch_n);
}

//Int24 c;
double readC(int ch_n, int* neg) {
  int8_t c_data[4] = {0};  // uint8_t c_data[4] = {0};
  int32_t c = 0; // uint32_t c = 0;
  int16_t msb = 0; // uint16_t msb = 0;
  int16_t lsb = 0;  // uint16_t lsb = 0;

  //Print c, msb and lsb to see how they are set up at the beginning
  Serial.println("C value initialized: ");
  Serial.println(c, BIN);
  Serial.println("");
  Serial.println("MSB: ");
  Serial.println(msb, BIN);
  Serial.println("");
  Serial.println("LSB: ");
  Serial.println(lsb, BIN);
  Serial.println("");
  
  // Read Measurement results on CIN1
  Wire.beginTransmission(SLAVE);
  Wire.write(0x00 + ch_n * 2);  
  Wire.endTransmission();

  Wire.requestFrom(SLAVE, 4);

  for (int i = 0; i < 4; i++) {
    c_data[i] = Wire.read();
  }

  //Printout the data acquired as single bytes element
  for (int i = 0; i < NUM_SENSORS; i++) {
    //Serial.print("CH" + String(i+5) + ": " + (data_list2[i], BIN) + " " );
    int bitsCount = sizeof( c_data[i] ) * 8;
    char str[ bitsCount + 1 ];
    uint8_t j = 0;
    while ( bitsCount-- )
      str[ j++ ] = bitRead( c_data[i], bitsCount ) + '0';
    str[ j ] = '\0';
    Serial.print("C_data[" + String(i+1) + "]:  ");
    Serial.println(str);
  }
  Serial.println("");

  msb = (c_data[0] << 8) + c_data[1];
  Serial.println("MSB: ");
  Serial.println(msb, BIN);
  Serial.println(""); 

  lsb = (c_data[2] << 8) + c_data[3];
  Serial.println("LSB: ");
  Serial.println(lsb, BIN);
  Serial.println(""); 

  //Convert byte array to int and print capacitance to serial, if trigger signal is high
  c = (msb << 8) + (lsb >> 9); 
  //Alternative:
  //c = (c_data[0] << 16) + (c_data[1] << 8) + c_data[2];  // MSB always first, LSB after
  
  //Printout the capacitnace value before the two's complement
  Serial.println("");
  Serial.println("24 bits Value before two's complement: ");
  Serial.println("C_value_before for channel " + String(ch_n) + ": ");
  int bitsCount = sizeof( c ) * 8;
  for (int i = (bitsCount-1); i >=0; i--) {
    if (c & (1 << i)) {
      Serial.print("1");
    } else {
      Serial.print("0");
    }
  }
  Serial.println("");

  if ((c & 0x80000000) != 0) { //if it is negative, I don't perform the two's complement
    // Serial.println("The value c[" + String(ch_n + 1) + "] is negative, NO two's compelment");
    // Serial.println("");
    c = (c & 0x00FFFFFF); //print  minus that value --> check if I need to bring in a one or not. It might be necessary to add a 1 in position [23] to represent it as negative
    (*neg) =1; //flag variable used to printing, passed as argument of this funciton
  }else if ((c & 0x80000000) == 0) { //if it is positive, I perform the two's complement    
    // Serial.println("The value c[" + String(ch_n + 1) + "] is positive, DO the two's compelment");
    // Serial.println("");
    c = (~c & 0x00FFFFFF) + 1;
    (*neg) =0; //flag variable used to printing, passed as argument of this funciton
  }  


  // Printout the capacitnace value after the two's complement
  Serial.println("24 bits Value after two's complement: ");
  Serial.println("C_value_after for channel " + String(ch_n) + ": ");
  for (int i = 31; i >= 0; i--) {
    if (c & (1 << i)) {
      Serial.print("1");
    } else {
      Serial.print("0");
    }
  }
  Serial.println("");

  //C decimal value
  Serial.println("C value decimal after two's complement for channel " + String(ch_n) + ": ");
  Serial.println(c);
  Serial.println("");

  double cap = (c / 524288.0) + Coffset;

  //Printout the converted cap value at the end of the code
  Serial.println("Cap value in binary: ");
  int* binaryPtr = reinterpret_cast<int*>(&cap);
  int binaryValue = *binaryPtr;
  int bitcountcap = sizeof(binaryValue) * 8;
  for (int i = (bitcountcap - 1); i >= 0; i--) {
    if (binaryValue & (1 << i)) {
      Serial.print("1");
    } else {
      Serial.print("0");
    }
  }
  Serial.println();
  Serial.println("");

  return cap;
}


